@@ -12,13 +12,10 @@
 #    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 #    License for the specific language governing permissions and limitations
 #    under the License.
-
-from __future__ import absolute_import
 import json as json_lib
 import logging
 import multiprocessing
 import os
-import six
 
 from cinder import coordination
 from cinder.db import api as db_api
@@ -31,7 +28,7 @@ cinder_objects.register_all()  # noqa
 from cinder.interface import util as cinder_interface_util
 from cinder import utils
 from cinder.volume import configuration
-from cinder.volume import manager
+from cinder.volume import manager  # noqa We need to import config options
 from oslo_config import cfg
 from oslo_log import log as oslo_logging
 from oslo_utils import importutils
@@ -70,13 +67,33 @@ class Backend(object):
     # With this dictionary the DB class can get the necessary data
     _volumes_inflight = {}
 
+    def __new__(cls, volume_backend_name, **driver_cfg):
+        # Prevent redefinition of an already initialized backend on the same
+        # persistence storage with a different configuration.
+        backend = Backend.backends.get(volume_backend_name)
+        if backend:
+            # If we are instantiating the same backend return the one we have
+            # saved (singleton pattern).
+            if driver_cfg == backend._original_driver_cfg:
+                return backend
+            raise ValueError('Backend named %s already exists with a different'
+                             ' configuration' % volume_backend_name)
+
+        return super(Backend, cls).__new__(cls)
+
     def __init__(self, volume_backend_name, **driver_cfg):
         if not self.global_initialization:
             self.global_setup()
+        # Instance already initialized
+        if volume_backend_name in Backend.backends:
+            return
+        # Save the original config before we add the backend name and template
+        # the values.
+        self._original_driver_cfg = driver_cfg.copy()
         driver_cfg['volume_backend_name'] = volume_backend_name
-        Backend.backends[volume_backend_name] = self
 
-        conf = self._set_backend_config(driver_cfg)
+        conf = self._get_backend_config(driver_cfg)
+        self._apply_backend_workarounds(conf)
         self.driver = importutils.import_object(
             conf.volume_driver,
             configuration=conf,
@@ -101,6 +118,7 @@ class Backend(object):
         self._stats = self._transform_legacy_stats(stats)
 
         self._pool_names = tuple(pool['pool_name'] for pool in stats['pools'])
+        Backend.backends[volume_backend_name] = self
 
     @property
     def pool_names(self):
@@ -190,164 +208,80 @@ class Backend(object):
         # the persistence plugin.
         db_api.IMPL = cls.persistence.db
 
-    # NOTE(geguileo): Staticmethod used instead of classmethod to make it work
-    # on Python3 when assigning the unbound method.
-    @staticmethod
-    def _config_parse(self):
-        """Replacer oslo_config.cfg.ConfigParser.parse for in-memory cfg."""
-        res = super(cfg.ConfigParser, self).parse(Backend._config_string_io)
-        return res
-
-    @classmethod
-    def _update_cinder_config(cls):
-        """Parse in-memory file to update OSLO configuration used by Cinder."""
-        cls._config_string_io.seek(0)
-        cls._parser.write(cls._config_string_io)
-
-        # Check if we have any multiopt
-        cls._config_string_io.seek(0)
-        current_cfg = cls._config_string_io.read()
-        if '\n\t' in current_cfg:
-            cls._config_string_io.seek(0)
-            cls._config_string_io.write(current_cfg.replace('\n\t', '\n'))
-
-        cls._config_string_io.seek(0)
-        cfg.CONF.reload_config_files()
-
     @classmethod
     def _set_cinder_config(cls, host, locks_path, cinder_config_params):
         """Setup the parser with all the known Cinder configuration."""
         cfg.CONF.set_default('state_path', os.getcwd())
         cfg.CONF.set_default('lock_path', '$state_path', 'oslo_concurrency')
-
-        cls._parser = six.moves.configparser.SafeConfigParser()
-        cls._parser.set('DEFAULT', 'enabled_backends', '')
+        cfg.CONF.version = cinderlib.__version__
 
         if locks_path:
-            cls._parser.add_section('oslo_concurrency')
-            cls._parser.set('oslo_concurrency', 'lock_path', locks_path)
-            cls._parser.add_section('coordination')
-            cls._parser.set('coordination',
-                            'backend_url',
-                            'file://' + locks_path)
+            cfg.CONF.oslo_concurrency.lock_path = locks_path
+            cfg.CONF.coordination.backend_url = 'file://' + locks_path
+
         if host:
-            cls._parser.set('DEFAULT', 'host', host)
+            cfg.CONF.host = host
 
-        # All other configuration options go into the DEFAULT section
-        cls.__set_parser_kv(cinder_config_params, 'DEFAULT')
-
-        # We replace the OSLO's default parser to read from a StringIO instead
-        # of reading from a file.
-        cls._config_string_io = six.moves.StringIO()
-        cfg.ConfigParser.parse = six.create_unbound_method(cls._config_parse,
-                                                           cfg.ConfigParser)
+        cls._validate_and_set_options(cinder_config_params)
 
         # Replace command line arg parser so we ignore caller's args
         cfg._CachedArgumentParser.parse_args = lambda *a, **kw: None
 
-        # Update the configuration with the options we have configured
-        cfg.CONF(project='cinder', version=cinderlib.__version__,
-                 default_config_files=['in_memory_file'])
-        cls._update_cinder_config()
-
-    @staticmethod
-    def __get_options_types(kvs):
-        """Get loaded oslo options and load driver if we know it."""
+    @classmethod
+    def _validate_and_set_options(cls, kvs, group=None):
+        """Validate options and substitute references."""
         # Dynamically loading the driver triggers adding the specific
         # configuration options to the backend_defaults section
-        if 'volume_driver' in kvs:
+        if kvs.get('volume_driver'):
             driver_ns = kvs['volume_driver'].rsplit('.', 1)[0]
             __import__(driver_ns)
-        opts = configuration.CONF._groups['backend_defaults']._opts
-        return opts
+            group = group or 'backend_defaults'
 
-    @classmethod
-    def __val_to_str(cls, val):
-        """Convert an oslo config value to its string representation.
+        for k, v in kvs.items():
+            try:
+                # set_override does the validation
+                cfg.CONF.set_override(k, v, group)
+            except cfg.NoSuchOptError:
+                # RBD keyring may be removed from the Cinder RBD driver, but
+                # the functionality will remain for cinderlib usage only, so we
+                # do the validation manually in that case.
+                # NOTE: Templating won't work on the rbd_keyring_conf, but it's
+                # unlikely to be needed.
+                if k == 'rbd_keyring_conf':
+                    if v and not isinstance(v, str):
+                        raise ValueError('%s must be a string' % k)
+                else:
+                    # Don't fail on unknown variables, behave like cinder
+                    LOG.warning('Unknown config option %s', k)
 
-        Lists and tuples are treated as ListOpt classes and converted to
-        "[value1,value2]" instead of the standard string representation of
-        "['value1','value2']".
+        oslo_group = getattr(cfg.CONF, str(group), cfg.CONF)
+        # Now that we have validated/templated everything set updated values
+        for k, v in kvs.items():
+            kvs[k] = getattr(oslo_group, k, v)
 
-        Dictionaries are treated as DictOpt and converted to 'k1:v1,k2:v2"
-        instead of the standard representation of "{'k1':'v1','k2':'v2'}".
+        # For global configuration we leave the overrides, but for drivers we
+        # don't to prevent cross-driver config polination.  The cfg will be
+        # set as an attribute of the configuration that's passed to the driver.
+        if group:
+            for k in kvs.keys():
+                try:
+                    cfg.CONF.clear_override(k, group, clear_cache=True)
+                except cfg.NoSuchOptError:
+                    pass
 
-        Anything else is converted to a string.
-        """
-        if isinstance(val, six.string_types):
-            return val
-
-        # Recursion is used to to handle options like u4p_failover_target that
-        # is a MultiOpt where each entry is a dictionary.
-        if isinstance(val, (list, tuple)):
-            return '[' + ','.join((cls.__val_to_str(v) for v in val)) + ']'
-
-        if isinstance(val, dict):
-            return ','.join('%s:%s' % (k, cls.__val_to_str(v))
-                            for k, v in val.items())
-
-        return six.text_type(val)
-
-    @classmethod
-    def __convert_option_to_string(cls, key, val, opts):
-        """Convert a Cinder driver cfg option into oslo config file format.
-
-        A single Python object represents multiple Oslo config types. For
-        example a list can be a ListOpt or a MultOpt, and their string
-        representations on a file are different.
-
-        This method uses the configuration option types to return the right
-        string representation.
-        """
-        opt = opts[key]['opt']
-
-        # Convert to a list for ListOpt opts were the caller didn't pass a list
-        if (isinstance(opt, cfg.ListOpt) and
-                not isinstance(val, (list, tuple))):
-            val = [val]
-
-        # For MultiOpt we need multiple entries in the file but ConfigParser
-        # doesn't support repeating the same entry multiple times, so we hack
-        # our way around it
-        elif isinstance(opt, cfg.MultiOpt):
-            if not isinstance(val, (list, tuple)):
-                val = [val] if val else []
-            val = [cls.__val_to_str(v) for v in val]
-            if not val:
-                val = ''
-            elif len(val) == 1:
-                val = val[0]
-            else:
-                val = (('%s\n' % val[0]) +
-                       '\n'.join('%s = %s' % (key, v) for v in val[1:]))
-
-        # This will handle DictOpt and ListOpt
-        if not isinstance(val, six.string_types):
-            val = cls.__val_to_str(val)
-        return val
-
-    @classmethod
-    def __set_parser_kv(cls, kvs, section):
-        """Set Oslo configuration options in our parser.
-
-        The Oslo parser needs to have the configuration options like they are
-        in a file, but we have them as Python objects, so we need to set them
-        for the parser in the format it is expecting them, strings.
-        """
-        opts = cls.__get_options_types(kvs)
-        for key, val in kvs.items():
-            string_val = cls.__convert_option_to_string(key, val, opts)
-            cls._parser.set(section, key, string_val)
-
-    def _set_backend_config(self, driver_cfg):
+    def _get_backend_config(self, driver_cfg):
+        # Create the group for the backend
         backend_name = driver_cfg['volume_backend_name']
-        self._parser.add_section(backend_name)
-        self.__set_parser_kv(driver_cfg, backend_name)
-        self._parser.set('DEFAULT', 'enabled_backends',
-                         ','.join(self.backends.keys()))
-        self._update_cinder_config()
-        config = configuration.Configuration(manager.volume_backend_opts,
-                                             config_group=backend_name)
+        cfg.CONF.register_group(cfg.OptGroup(backend_name))
+
+        # Validate and set config options
+        self._validate_and_set_options(driver_cfg)
+        backend_group = getattr(cfg.CONF, backend_name)
+        for key, value in driver_cfg.items():
+            setattr(backend_group, key, value)
+
+        # Return the Configuration that will be passed to the driver
+        config = configuration.Configuration([], config_group=backend_name)
         return config
 
     @classmethod
@@ -384,6 +318,14 @@ class Backend(object):
         cls.global_initialization = True
         cls.output_all_backend_info = output_all_backend_info
 
+    def _apply_backend_workarounds(self, config):
+        """Apply workarounds for drivers that do bad stuff."""
+        if 'netapp' in config.volume_driver:
+            # Workaround NetApp's weird replication stuff that makes it reload
+            # config sections in get_backend_configuration.  OK since we don't
+            # support replication.
+            cfg.CONF.list_all_sections = lambda: config.volume_backend_name
+
     @classmethod
     def _set_logging(cls, disable_logs):
         if disable_logs:
@@ -460,38 +402,87 @@ class Backend(object):
             self.volumes
 
     @staticmethod
-    def list_supported_drivers():
-        """Returns dictionary with driver classes names as keys."""
+    def list_supported_drivers(output_version=1):
+        """Returns dictionary with driver classes names as keys.
 
-        def convert_oslo_config(oslo_options):
-            options = []
-            for opt in oslo_options:
-                tmp_dict = {k: str(v) for k, v in vars(opt).items()
-                            if not k.startswith('_')}
-                options.append(tmp_dict)
-            return options
+        The output of the method changes from version to version, so we can
+        pass the output_version parameter to specify which version we are
+        expecting.
 
-        def list_drivers(queue):
+        Version 1: Original output intended for human consumption, where all
+                   dictionary values are strings.
+        Version 2: Improved version intended for automated consumption.
+                   - type is now a dictionary with detailed information
+                   - Values retain their types, so we'll no longer get 'None'
+                      or 'False'.
+        """
+        def get_vars(obj):
+            return {k: v for k, v in vars(obj).items()
+                    if not k.startswith('_')}
+
+        def get_strs(obj):
+            return {k: str(v) for k, v in vars(obj).items()
+                    if not k.startswith('_')}
+
+        def convert_oslo_config(oslo_option, output_version):
+            if output_version != 2:
+                return get_strs(oslo_option)
+
+            res = get_vars(oslo_option)
+            type_class = res['type']
+            res['type'] = get_vars(oslo_option.type)
+            res['type']['type_class'] = type_class
+            return res
+
+        def fix_cinderlib_options(driver_dict, output_version):
+            # The rbd_keyring_conf option is deprecated and will be removed for
+            # Cinder, because it's a security vulnerability there (OSSN-0085),
+            # but it isn't for cinderlib, since the user of the library already
+            # has access to all the credentials, and cinderlib needs it to work
+            # with RBD, so we need to make sure that the config option is
+            # there whether it's reported as deprecated or removed from Cinder.
+            RBD_KEYRING_CONF = cfg.StrOpt('rbd_keyring_conf',
+                                          default='',
+                                          help='Path to the ceph keyring file')
+
+            if driver_dict['class_name'] != 'RBDDriver':
+                return
+            rbd_opt = convert_oslo_config(RBD_KEYRING_CONF, output_version)
+            for opt in driver_dict['driver_options']:
+                if opt['dest'] == 'rbd_keyring_conf':
+                    opt.clear()
+                    opt.update(rbd_opt)
+                    break
+            else:
+                driver_dict['driver_options'].append(rbd_opt)
+
+        def list_drivers(queue, output_version):
             cwd = os.getcwd()
             # Go to the parent directory directory where Cinder is installed
             os.chdir(utils.__file__.rsplit(os.sep, 2)[0])
             try:
                 drivers = cinder_interface_util.get_volume_drivers()
                 mapping = {d.class_name: vars(d) for d in drivers}
-                # Drivers contain class instances which are not serializable
                 for driver in mapping.values():
                     driver.pop('cls', None)
                     if 'driver_options' in driver:
-                        driver['driver_options'] = convert_oslo_config(
-                            driver['driver_options'])
+                        driver['driver_options'] = [
+                            convert_oslo_config(opt, output_version)
+                            for opt in driver['driver_options']
+                        ]
+                        fix_cinderlib_options(driver, output_version)
             finally:
                 os.chdir(cwd)
             queue.put(mapping)
 
+        if not (1 <= output_version <= 2):
+            raise ValueError('Acceptable versions are 1 and 2')
+
         # Use a different process to avoid having all driver classes loaded in
         # memory during our execution.
         queue = multiprocessing.Queue()
-        p = multiprocessing.Process(target=list_drivers, args=(queue,))
+        p = multiprocessing.Process(target=list_drivers,
+                                    args=(queue, output_version))
         p.start()
         result = queue.get()
         p.join()
